/* Polynomial Adddtion Using Generalised Linked Lists */
/* -------------------------------------------------- */

/* Including Header Files */
/* ---------------------- */

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <alloc.h>
#include <ctype.h>

/* Defining Global Variables */
/* ------------------------- */

typedef struct gll
{
	int coefficent;
	int power;
	char variable;
	struct gll *down;
	struct gll *next;
}	GLL;

GLL *first,*second,*result;
int prev_exp_parsed=0;
GLL *root=NULL;

/* Defining Function Prototypes */
/* ---------------------------- */

int islegal(char ch);
char *getstring(void);
GLL *newnode(void);
GLL *parse(char *polynomial);
void print_polynomial(GLL *first);
char *extract(char *polynomial,int *index);
void add_polynomial(GLL *first ,GLL *second);
int ismatch(GLL *first, GLL *second);
void error(void);

/* Main Manages Inputs & Outpuits */
/* ------------------------------ */

void main(void)
{
	char *first_polynomial,*second_polynomial;
	clrscr();
	printf("\n\n\tPolynomial Addition Using Generalised Linked List");
	printf("\n\t-------------------------------------------------");
	printf("\n\n\tEnter The First Polynomial: ");
	first_polynomial=getstring();
	printf("\n\n\tEnter The Second Polynomial: ");
	second_polynomial=getstring();

	first=parse(first_polynomial);
	prev_exp_parsed=0;
	second=parse(second_polynomial);
	result=first;
	first=first->next;
	result->next=NULL;
	add_polynomial(result,first);
	add_polynomial(result,second);
	print_polynomial(result);
}

/* Checks For Illegal Characters */
/* ----------------------------- */

int islegal(char ch)
{
	if(ch=='+' || ch=='^'|| ch=='\r' || isalnum(ch))
		return 1;
	return 0;
}

/* Takes The String From The User */
/* ------------------------------ */

char *getstring(void)
{
	int index=-1;
	char *string=(char *)malloc(100);
	do
	{
		index++;
		do
			*(string+index)=getch();
		while(! (islegal(*(string+index))) );
		putch(*(string+index));
	}while(*(string+index)!='\r');

	*(string+index)='\0';
	return(string);
}

/* Creates A New Node */
/* ------------------ */

GLL *newnode(void)
{
	GLL *temp=(GLL *)malloc(sizeof(GLL));
	temp->down=temp->next=NULL;
	return(temp);
}

/* Converts String Polynomial To Linked List Polynomial */
/* ---------------------------------------------------- */

GLL *parse(char *polynomial)
{
	GLL *temp=newnode();
	int index=0, coefficent=1,power=1,count;
	int variable_found=0;
	char ans[5];
	char *intermediate;

	while((*(polynomial+index))!='\0'&& !(prev_exp_parsed))
	{
		if(isdigit(*(polynomial+index)) )
		{
			count=0;
			while(isdigit(*(polynomial+index)))
			{
				ans[count++]=*(polynomial+index);
				index++;
			}
			if(count > 4)
				count=4;
			ans[count]='\0';
			coefficent=atoi(ans);
		}
		if(isalpha(*(polynomial+index)))
		{
			if(!(variable_found))
			{
				temp->variable=*(polynomial+index);
				index++;
				variable_found=1;
			}
			else
			{
				intermediate=extract(polynomial,&index);
				temp->down=parse(intermediate);
			}
		}
		if(*(polynomial+index)=='^')
		{
			if(!(variable_found))
				error();
			index++;
			if(isdigit(*(polynomial+index)) )
			{
				count=0;
				while(isdigit(*(polynomial+index)))
				{
					ans[count++]=*(polynomial+index);
					index++;
				}
				if(count < 5)
					count=4;
				ans[count]='\0';
				power=atoi(ans);
			}
		}
		if(*(polynomial+index)=='+')
		{
			index++;
			temp->coefficent=coefficent;
			temp->power=power;
			polynomial=polynomial+index;
			temp->next=parse(polynomial);
			prev_exp_parsed=1;
		}
	}
	temp->coefficent=coefficent;
	temp->power=power;

	return(temp);
}

/* Print The Polynomial From Generalised Linked List */
/* ------------------------------------------------- */

void print_polynomial(GLL *first)
{
	int index=0,count=0,head_encountered=0;
	char ans[5];
	GLL *temp1=first;
	char *polynomial=(char *)calloc(100,sizeof(char));

	while(temp1!=NULL)
	{
		first=temp1;
		while(first!=NULL)
		{
			if(head_encountered==0)
			{
				count=0;
				if(first->coefficent!=1)
				{
					itoa(first->coefficent,ans,10);
					while(ans[count]!='\0')
					{
						*(polynomial+index)=ans[count];
						count++;index++;
					}
				}
				head_encountered=1;
			}
			*(polynomial+index)=first->variable;
			index++;
			if(first->power!=1)
			{
				*(polynomial+index)='^';
				index++;
				count=0;
				itoa(first->power,ans,10);
				while(ans[count]!='\0')
				{
					*(polynomial+index)=ans[count];
					count++;index++;
				}
			}
			first=first->down;
		}
		temp1=temp1->next;
		head_encountered=0;
		if(temp1!=NULL)
		{
			*(polynomial+index)='+';
			index++;
		}

	}
	*(polynomial+index)='\0';
	printf("\n\n\tThe Final Answer Is %s",polynomial);
	getch();
}

/* The Extract Will Create A Subterm Of The Polynomial */
/* --------------------------------------------------- */

char *extract(char *polynomial,int *index)
{
	int count=0;
	char *new_polynomial=(char *)malloc(100);
	while( *(polynomial+(*index))!='+' && *(polynomial+(*index))!='\0')
	{
		*(new_polynomial+count)=*(polynomial+(*index));
		(*index)++;
		count++;
	}
	*(new_polynomial+(count))='\0';
	return(new_polynomial);
}

/* This Function Will Add The Two Linked List Polynomial */
/* ----------------------------------------------------- */

void add_polynomial(GLL *first,GLL *second)
{
	int match;
	GLL *temp,*temp1,*temp2;
	temp2=second;
	while(temp2!=NULL)
	{
		temp=first;
		match=0;
		while(temp!=NULL && match==0)
		{
			match=ismatch(temp,temp2);
			if(match==0)
			{
				temp1=temp;
				temp=temp->next;
			}
		}
		if(match==0)
		{
			temp1->next=temp2;
			temp2=temp2->next;
			temp1->next->next=NULL;
		}
		else
		{
			temp->coefficent+=temp2->coefficent;
			temp2=temp2->next;
		}
	}
}

/* This Will Check Weather Two Terms Are Of Same Degrees */
/* ----------------------------------------------------- */

int ismatch(GLL *first, GLL *second)
{
	while(first!=NULL && second!=NULL)
	{
		if(first->power==second->power
		   &&
		   first->variable==second->variable)
		{
			first=first->down;
			second=second->down;
		}
		else
			return 0;
	}
	if(first==NULL && second==NULL)
		return 1;
	else
		return 0;
}

/* This Function Will Show Error On Invalid Polynomial */
/* --------------------------------------------------- */

void error(void)
{
	printf("\n\n\t\aERROR!!! ");
	printf("The Polynomial Is Not In Requested Form");
	printf("\n\tThe Program Will Be Terminated...");
	getch();
	exit(1);
}